/**
 * @file linmaster.c
 *
 * @brief This file contains LIN bus master routines for the USBlini firmware
 *        project.
 *
 * @author Thomas Fischl (https://www.fischl.de/contact.html)
 * @copyright (c) 2021 Thomas Fischl
 * 
 * This file is part of the USBlini firmware project and is copyrighted by the
 * authors listed above. It is free for private or educational non-commercial use
 * (for a commercial license please contact the authors). It may not be
 * redistributed without the prior written consent of the authors.
 *
 * It is provided "as is" without warranty of any kind, either express or implied,
 * including without limitation any implied warranties of condition, uninterrupted
 * use, merchantability, fitness for a particular purpose, or non-infringement. In
 * no event shall the authors be liable for any direct or indirect damages arising
 * in any way out of the use of it.
 *
 */

#include "usblini.h"
#include "lin.h"
#include "clock.h"
#include "linslave.h"
#include "linmaster.h"

#define SEQUENCE_SIZE 16

typedef struct
{
    uint16_t frametime;
    uint8_t id;
} mastersequence_t;

mastersequence_t masterseq[SEQUENCE_SIZE];
uint8_t masterseq_length = 0;
uint8_t masterseq_pos;
uint16_t masterseq_period = 0;
uint16_t timestamp;
volatile uint16_t masterseq_frametimeout = 0;
volatile uint16_t masterseq_periodtimeout = 0;

/**
 * @brief Send given frame (single write)
 * @param id Identifier of message
 * @param data Data of message
 * @param length Data length
 * @param checksum_mode Checksum mode (see LIN_CHECKSUM_*)
 * @param source Source (see REPORT_SOURCE_*)
 */
void linmaster_sendFrame(uint8_t id, uint8_t * data, uint8_t length, uint8_t checksum_mode, uint8_t source) {
   
    // prepare checksum
    uint16_t checksum = (checksum_mode == LIN_CHECKSUM_LIN2 ? lin_getPID(id) : 0);
    
    // calculate checksum
    uint8_t i;
    for (i = 0; i < length; i++) {
        checksum += data[i];
        if (checksum > 0xff) checksum -= 0xff;
    }
    
    // add checksum
    if (length && (checksum_mode != LIN_CHECKSUM_NONE)) {
        checksum = ~checksum;
        data[length++] = (uint8_t)checksum;
    }
    
    lin_sendDataWithHeader(id, data, length, source);
}

/**
 * @brief Set given master sequence. Given identifiers are requested periodically
 * @param period Period time of complete sequence.
 * @param frametime Time of one frame (one id)
 * @param ids List of identifiers
 * @param length Length of ID list
 */
void linmaster_setSequence(uint16_t period, uint16_t frametime, uint8_t * ids, uint8_t length) {
        
    masterseq_length = 0;
    
    uint8_t i;
    for (i = 0; i < length; i++) {
        masterseq[i].id = ids[i];
        masterseq[i].frametime = frametime;
    }
    
    masterseq_period = period;
    masterseq_length = length;
    masterseq_pos = 0;
}

/**
 * @brief Clear master sequence
 */
void linmaster_clearSequence(void) {
    masterseq_length = 0;    
}

/**
 * @brief Do module processing
 */
void linmaster_process(void) {
    
    uint16_t frametimeout;
    uint16_t periodtimeout;
    
    if (masterseq_length == 0) return;
    if (lin_getReceiveState() != RECEIVESTATE_IDLE) return;
    if (lin_getSendState() != SENDSTATE_IDLE) return;
    
    clock_disableISR(); periodtimeout = masterseq_periodtimeout; clock_enableISR();
    if ((masterseq_pos == 0) && (periodtimeout > 0)) return;  // wait for end of last period

    clock_disableISR(); frametimeout = masterseq_frametimeout; clock_enableISR();
    if (frametimeout > 0) return; // wait for end of last frame
    
    clock_disableISR();
    masterseq_frametimeout = masterseq[masterseq_pos].frametime;
    if (masterseq_pos == 0) {
        masterseq_periodtimeout = masterseq_period;
    }
    clock_enableISR();
    
    lin_sendHeader(masterseq[masterseq_pos].id, REPORT_SOURCE_MASTER);
        
    masterseq_pos++;
    if (masterseq_pos >= masterseq_length) {
        masterseq_pos = 0;            
    }        
    
}

/**
 * @brief LIN master interrupt function
 */
inline void linmaster_timer_isr(void) {
    if (masterseq_frametimeout) masterseq_frametimeout--;
    if (masterseq_periodtimeout) masterseq_periodtimeout--;
}