/**
 * @file linslave.c
 *
 * @brief This file contains LIN bus slave routines for the USBlini firmware
 *        project.
 *
 * @author Thomas Fischl (https://www.fischl.de/contact.html)
 * @copyright (c) 2021 Thomas Fischl
 * 
 * This file is part of the USBlini firmware project and is copyrighted by the
 * authors listed above. It is free for private or educational non-commercial use
 * (for a commercial license please contact the authors). It may not be
 * redistributed without the prior written consent of the authors.
 *
 * It is provided "as is" without warranty of any kind, either express or implied,
 * including without limitation any implied warranties of condition, uninterrupted
 * use, merchantability, fitness for a particular purpose, or non-infringement. In
 * no event shall the authors be liable for any direct or indirect damages arising
 * in any way out of the use of it.
 *
 */

#include "usblini.h"
#include "lin.h"
#include "statusreport.h"
#include "linslave.h"

typedef struct
{
    uint8_t active;
    uint8_t id;
    uint8_t length;
    uint8_t data[SLAVETABLE_DATA_SIZE];
    uint16_t reloadvalue;
    uint16_t counter;
    uint16_t resetmask;
    
} slavetable_t;

slavetable_t slavetable[SLAVETABLE_SIZE];

uint8_t frameid = HEADERID_NONE;

/**
 * @brief Initialize LIN slave module
 */
void linslave_init(void) {
    uint8_t i;
    for (i = 0; i < SLAVETABLE_SIZE; i++) {
        slavetable[i].active = 0;
        slavetable[i].length = 0;
    }
}

/**
 * @brief Set given slave table item.
 * @param tableid Table id
 * @param id Identifier of message
 * @param data Data of message
 * @param length Length of data
 * @param checksum_mode Checksum mode (see LIN_CHECKSUM_*)
 * @param reloadvalue Value which is set at table item reset
 * @param resetmask Mask of table items to reset at counter value zero
 */
void linslave_setTableItem(uint8_t tableid, uint8_t id, uint8_t * data, uint8_t length, uint8_t checksum_mode, uint16_t reloadvalue, uint16_t resetmask) {

    slavetable[tableid].length = length;
    
    if (length == 0) {
        slavetable[tableid].active = 0;
        statusreport_setSlaveTableStatus(tableid, 0);
        return;
    }
    
    slavetable[tableid].id = id;
    slavetable[tableid].active = 1;
    statusreport_setSlaveTableStatus(tableid, 1);
    slavetable[tableid].reloadvalue = reloadvalue;
    slavetable[tableid].counter = reloadvalue;
    slavetable[tableid].resetmask = resetmask;    
    
    // prepare checksum
    uint16_t checksum = (checksum_mode == LIN_CHECKSUM_LIN2 ? lin_getPID(id) : 0);
    
    // set data bytes
    uint8_t i;
    for (i = 0; i < length; i++) {
        slavetable[tableid].data[i] = data[i];
        checksum += data[i];
        if (checksum > 0xff) checksum -= 0xff;
    }
    
    // set checksum
    if (length && (checksum_mode != LIN_CHECKSUM_NONE)) {
        checksum = ~checksum;
        slavetable[tableid].data[length] = (uint8_t)checksum;
        slavetable[tableid].length++;
    }
}

/**
 * @param Set given reload value for reset condition
 * @param tableid Table id
 * @param reloadvalue Reload value
 */
void linslave_setReloadvalue(uint8_t tableid, uint16_t reloadvalue) {
    slavetable[tableid].reloadvalue = reloadvalue;   
    slavetable[tableid].counter = reloadvalue;
}

/**
 * @brief Set given reset mask. If counter reaches zero, table items in reset
 * mask are resetted.
 * @param tableid Table id
 * @param resetmask Reset mask
 */
void linslave_setResetmask(uint8_t tableid, uint16_t resetmask) {
    slavetable[tableid].resetmask = resetmask;
}

/**
 * @brief Trigger slave processing of given PID
 * @param pid Protected id of message to process
 */
void linslave_triggerIDProcessing(uint8_t pid) {
    frameid = pid & 0x03f;
}

/**
 * @brief Reset given table item
 * @param tableid Table id
 */
void linslave_reset(uint8_t tableid) {
    
    slavetable[tableid].counter = slavetable[tableid].reloadvalue;
    slavetable[tableid].active = (slavetable[tableid].length > 0);
    statusreport_setSlaveTableStatus(tableid, slavetable[tableid].active);
}

/**
 * @brief Apply given reset mask
 * @param resetmask Reset bit mask
 */
void linslave_applyResetMask(uint16_t resetmask) {
    uint8_t i; 
    
    for (i = 0; i < SLAVETABLE_SIZE; i++) {
        if (resetmask & 1) {
            linslave_reset(i);
        }
        resetmask = resetmask >> 1;
    }    
}

/**
 * @brief Send out message at given table row
 * @param tableid Table id/row
 */
void linslave_send(uint8_t tableid) {
    
    lin_sendData(slavetable[tableid].data, slavetable[tableid].length, REPORT_SOURCE_SLAVE);
    
    if (slavetable[tableid].counter) {
        slavetable[tableid].counter --;
            
        if (slavetable[tableid].counter == 0) {

            // counter zero -> do action                
            slavetable[tableid].active = 0;
            statusreport_setSlaveTableStatus(tableid, slavetable[tableid].active);
            linslave_applyResetMask(slavetable[tableid].resetmask);
        }
    }
}

/**
 * @brief Do module processing
 */
void linslave_process(void) {
    
    // check if there is an id pending for processing
    if (frameid == HEADERID_NONE) return;

    // go through slave table and send out data on first match
    uint8_t i;
    for (i = 0; i < SLAVETABLE_SIZE; i++) {
        if (slavetable[i].active && (slavetable[i].id == frameid)) {
            linslave_send(i);
            break;
        }
    }
    
    // mark as processed
    frameid = HEADERID_NONE;    
}
