/**
 * @file usb_vendor.c
 *
 * @brief This file contains USB vendor routines for the USBlini firmware
 *        project.
 *
 * @author Thomas Fischl (https://www.fischl.de/contact.html)
 * @copyright (c) 2021 Thomas Fischl
 * 
 * This file is part of the USBlini firmware project and is copyrighted by the
 * authors listed above. It is free for private or educational non-commercial use
 * (for a commercial license please contact the authors). It may not be
 * redistributed without the prior written consent of the authors.
 *
 * It is provided "as is" without warranty of any kind, either express or implied,
 * including without limitation any implied warranties of condition, uninterrupted
 * use, merchantability, fitness for a particular purpose, or non-infringement. In
 * no event shall the authors be liable for any direct or indirect damages arising
 * in any way out of the use of it.
 *
 */
#include "usblini.h"
#include "lin.h"
#include "linmaster.h"
#include "linslave.h"
#include "statusreport.h"
#include "bootloader.h"

#define CMD_ECHO                  0x01
#define CMD_START_BOOTLOADER      0x02
#define CMD_RESET                 0x03
#define CMD_SET_BAUDRATE          0x04
#define CMD_CLEAR_ERRORFLAGS      0x05

#define CMD_MASTER_WRITE          0x11
#define CMD_MASTER_SET_SEQUENCE   0x12 

#define CMD_SLAVE_SET_FRAME       0x21
#define CMD_SLAVE_SET_RELOADVALUE 0x22 
#define CMD_SLAVE_SET_RESETMASK   0x23

#define MAX_DATA 9
#define MAX_SEQUENCE_SIZE 16

#define OUTTRANSFER_STATE_IDLE 0
#define OUTTRANSFER_STATE_MASTER_WRITE 1
#define OUTTRANSFER_STATE_MASTER_SET_SEQUENCE 2
#define OUTTRANSFER_STATE_SLAVE_SET_FRAME 3

typedef struct
{
    uint8_t id;
    uint8_t pos;
    uint8_t checksum_mode;
    uint8_t length;
    uint8_t data[MAX_DATA];
} cmd_master_write_t;

typedef struct
{
    uint16_t period;
    uint16_t frametime;
    uint8_t pos;
    uint8_t length;
    uint8_t ids[MAX_SEQUENCE_SIZE];
} cmd_master_set_sequence_t;

typedef struct
{
    uint8_t id;
    uint8_t pos;
    uint8_t checksum_mode;
    uint8_t length;
    uint8_t data[MAX_DATA];
    uint8_t tableid;
} cmd_slave_set_frame_t;

cmd_master_write_t cmd_master_write;
cmd_master_set_sequence_t cmd_master_set_sequence;
cmd_slave_set_frame_t cmd_slave_set_frame;
uint8_t expected = 0;
uint8_t outtransfer_state = OUTTRANSFER_STATE_IDLE;
uint8_t responsedata[MAX_DATA];
uint8_t responsedata_pos;
uint8_t responsedata_length;


/**
 * @brief Handle the USB setup packets
 * 
 * @param data USB packet data for request
 * @param response USB packet data for response
 * @return Length of data to send out
 */
uint8_t usb_vendor_setup(uint8_t * data, uint8_t * response) {

    switch (data[1]) {

        case CMD_ECHO: // echo (for transfer reliability testing)
            response[0] = data[2];
            response[1] = data[3];
            return 2;
            
        case CMD_START_BOOTLOADER:
            if ((data[2] == 0x37) && (data[3] == 0x52))
                bootloader_triggerStart();
            return 0;
            
        case CMD_RESET:
            lin_init();
            linmaster_setSequence(0, 0, 0, 0);
            linslave_init();
            statusreport_reset();
            return 0;
            
        case CMD_SET_BAUDRATE:
            lin_setBaudrate((uint16_t) data[2] + ((uint16_t) data[3] << 8));
            lin_setAutobaud(data[4]);
            return 0;          
                        
        case CMD_CLEAR_ERRORFLAGS:
            statusreport_clearErrorflags(data[2]);
            return 0;
            
        case CMD_MASTER_WRITE:
            
            cmd_master_write.id = data[2];
            cmd_master_write.checksum_mode = data[3];
            cmd_master_write.length = data[6];                        
            if (cmd_master_write.length > MAX_DATA) cmd_master_write.length = MAX_DATA;
            expected = cmd_master_write.length;
            cmd_master_write.pos = 0;
            
            if (expected == 0) {
                linmaster_sendFrame(cmd_master_write.id, cmd_master_write.data, cmd_master_write.length, LIN_CHECKSUM_NONE, REPORT_SOURCE_USER);
                outtransfer_state = OUTTRANSFER_STATE_IDLE;
            } else {
                outtransfer_state = OUTTRANSFER_STATE_MASTER_WRITE;
            }
            
            return (cmd_master_write.length ? 0xff : 0x00);
            
        case CMD_MASTER_SET_SEQUENCE:
            
            cmd_master_set_sequence.period = (uint16_t) data[2] + ((uint16_t) data[3] << 8);
            cmd_master_set_sequence.frametime = (uint16_t) data[4] + ((uint16_t) data[5] << 8);
            cmd_master_set_sequence.length = data[6];
            if (cmd_master_set_sequence.length > MAX_SEQUENCE_SIZE) cmd_master_set_sequence.length = MAX_SEQUENCE_SIZE;
            expected = cmd_master_set_sequence.length;
            cmd_master_set_sequence.pos = 0;
            
            if (expected == 0) {
                linmaster_clearSequence();
                outtransfer_state = OUTTRANSFER_STATE_IDLE;
            } else {
                outtransfer_state = OUTTRANSFER_STATE_MASTER_SET_SEQUENCE;
            }
            
            return (cmd_master_set_sequence.length ? 0xff : 0x00);
            
        case CMD_SLAVE_SET_FRAME:
            
            cmd_slave_set_frame.id = data[2];
            cmd_slave_set_frame.checksum_mode = data[3];
            cmd_slave_set_frame.tableid = data[4];
            cmd_slave_set_frame.length = data[6];
            if (cmd_slave_set_frame.length > MAX_DATA) cmd_slave_set_frame.length = MAX_DATA;
            expected = cmd_slave_set_frame.length;
            cmd_slave_set_frame.pos = 0;
            
            if (expected == 0) {
                linslave_setTableItem(cmd_slave_set_frame.tableid, cmd_slave_set_frame.id, cmd_slave_set_frame.data, cmd_slave_set_frame.length, cmd_slave_set_frame.checksum_mode, 0, 0);
                outtransfer_state = OUTTRANSFER_STATE_IDLE;
            } else {
                outtransfer_state = OUTTRANSFER_STATE_SLAVE_SET_FRAME;
            }
            
            return (cmd_slave_set_frame.length ? 0xff : 0x00);
            
        case CMD_SLAVE_SET_RELOADVALUE:            
            linslave_setReloadvalue(data[4], (uint16_t) data[2] + ((uint16_t) data[3] << 8));            
            return 0;
            
        case CMD_SLAVE_SET_RESETMASK:            
            linslave_setResetmask(data[4], (uint16_t) data[2] + ((uint16_t) data[3] << 8));            
            return 0;            
    }

    return 0;
}

/**
 * @brief Handle USB in requests
 * 
 * @param data Pointer to packet data to fill
 * @param length Count of maximum data length
 * @return Count of written data
 */
uint8_t usb_vendor_in(uint8_t * data, uint8_t length) {

    uint8_t i;

    if (length > responsedata_length) length = responsedata_length;
    
    for (i = 0; i < length; i++) {
        data[i] = responsedata[responsedata_pos++];        
        responsedata_length--;
    }

    return length;
}

/**
 * @brief Handle USB out requests
 * 
 * @param data Byte field containing USB data packet
 * @param length Length of data packet
 * @return 1 if ack should be sent
 */
uint8_t usb_vendor_out(uint8_t * data, uint8_t length) {

    uint8_t i;
    
    if (!expected) return 0;
    
    switch (outtransfer_state) {
        case OUTTRANSFER_STATE_MASTER_WRITE:
            
            i = 0;
            while ((length > 0) && (expected > 0)) {
                cmd_master_write.data[cmd_master_write.pos++] = data[i++];
                expected--;
                length--;
            }
            
            if (expected == 0) {
                linmaster_sendFrame(cmd_master_write.id, cmd_master_write.data, cmd_master_write.length, cmd_master_write.checksum_mode, REPORT_SOURCE_USER); 
            }
            
            return 1;
            
        case OUTTRANSFER_STATE_MASTER_SET_SEQUENCE:
            
            i = 0;
            while ((length > 0) && (expected > 0)) {
                cmd_master_set_sequence.ids[cmd_master_set_sequence.pos++] = data[i++];
                expected--;
                length--;
            }
            
            if (expected == 0) {
                linmaster_setSequence(cmd_master_set_sequence.period, cmd_master_set_sequence.frametime, cmd_master_set_sequence.ids, cmd_master_set_sequence.length);
            }
            return 1;
            
        case OUTTRANSFER_STATE_SLAVE_SET_FRAME:
            
            i = 0;
            while ((length > 0) && (expected > 0)) {
                cmd_slave_set_frame.data[cmd_slave_set_frame.pos++] = data[i++];
                expected--;
                length--;
            }
            
            if (expected == 0) {
                linslave_setTableItem(cmd_slave_set_frame.tableid, cmd_slave_set_frame.id, cmd_slave_set_frame.data, cmd_slave_set_frame.length, cmd_slave_set_frame.checksum_mode, 0, 0);
            }
            
            return 1;
            
        default:
            return 0;            
    }   

    return 1;
}

